/*
Interfacing the Arduino Duemilanove to Am29F010 devices.

Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

const int CE = A5, OE = A4, WE = A3, A16 = A2,
#ifdef CIRCUIT_74HC273
          CS1 = 2, CS2 = 3,                     /* using 74HC273 */
#endif
#ifdef CIRCUIT_74HC595
          SHIFT_SCK = 2, SHIFT_RCK = 3,         /* using 74HC595 */
          SHIFT_SI = 12,
#endif
          DQ0 = 7, DQ1 = 6, DQ2 = 5, DQ3 = 4,
          DQ4 = 8, DQ5 = 9, DQ6 = 10, DQ7 = 11;

const int datapins[8] = {DQ7, DQ6, DQ5, DQ4, DQ3, DQ2, DQ1, DQ0};
const int BUFSIZE = 9;
char inbuffer[BUFSIZE];
int nread = 0;

const char RESET[][9] = {
    "W05555AA", "W02AAA55", "W05555F0"
    };
const char ERASE[][9] = {
    "W05555AA", "W02AAA55", "W0555580", "W05555AA", "W02AAA55"
    };
const char PROGRAM[][9] = {
    "W05555AA", "W02AAA55", "W05555A0"
    };

int fromHex(char c)
{
    if ((c >= 48) && (c <= 57))
        return c - 48;
    if ((c >= 65) && (c <= 70))
        return c - 65 + 10;
    if ((c >= 97) && (c <= 102))
        return c - 97 + 10;
    return 0;
}

void setDataOut()
{
    pinMode(DQ0, OUTPUT);
    pinMode(DQ1, OUTPUT);
    pinMode(DQ2, OUTPUT);
    pinMode(DQ3, OUTPUT);
    pinMode(DQ4, OUTPUT);
    pinMode(DQ5, OUTPUT);
    pinMode(DQ6, OUTPUT);
    pinMode(DQ7, OUTPUT);
}

void setDataIn()
{
    pinMode(DQ0, INPUT);
    pinMode(DQ1, INPUT);
    pinMode(DQ2, INPUT);
    pinMode(DQ3, INPUT);
    pinMode(DQ4, INPUT);
    pinMode(DQ5, INPUT);
    pinMode(DQ6, INPUT);
    pinMode(DQ7, INPUT);
}

int sampleData()
{
    int i = 0; /* start at DQ7 */
    int data = 0;

    while (i < 8)
    {
        data <<= 1;
        data += digitalRead(datapins[i]);
        i++;
    }

    return data;
}

int readData()
{
    int data = 0;

    setDataIn();
    delayMicroseconds(1);
    data = sampleData();
    setDataOut();
    return data;
}

void writeData(int data)
{
    int i = 8; /* start at DQ0 */

    while (i > 0)
    {
        i--;
        digitalWrite(datapins[i], data % 2);
        data >>= 1;
    }
}

#ifdef CIRCUIT_74HC273
void setAddress(int high, int low)
{
    /* The top bit is sent directly. */

    if (high & 0x100)
        digitalWrite(A16, HIGH);
    else
        digitalWrite(A16, LOW);

    /* The lower 16 bits are sent via the latches. */

    writeData(high);
    delayMicroseconds(1);
    digitalWrite(CS2, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS2, LOW);
    writeData(low);
    delayMicroseconds(1);
    digitalWrite(CS1, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS1, LOW);
}
#endif

#ifdef CIRCUIT_74HC595
void writeAddress(int v)
{
    int mask;

    for (mask = 0x80; mask; mask >>= 1)
    {
        if (v & mask)
        {
            digitalWrite(SHIFT_SI, HIGH);
        }
        else
        {
            digitalWrite(SHIFT_SI, LOW);
        }

        delayMicroseconds(1);
        digitalWrite(SHIFT_SCK, HIGH);
        delayMicroseconds(1);
        digitalWrite(SHIFT_SCK, LOW);
    }

    delayMicroseconds(1);
    digitalWrite(SHIFT_RCK, HIGH);
    delayMicroseconds(1);
    digitalWrite(SHIFT_RCK, LOW);
}

void setAddress(int high, int low)
{
    /* The top bit is sent directly. */

    if (high & 0x100)
        digitalWrite(A16, HIGH);
    else
        digitalWrite(A16, LOW);

    /* The lower 16 bits are sent via the shift registers. */

    writeAddress(high);
    writeAddress(low);
}
#endif

int readOp(int high, int low)
{
    int data;

    digitalWrite(CE, LOW);
    setAddress(high, low);
    delayMicroseconds(1);
    digitalWrite(OE, LOW);
    data = readData();
    delayMicroseconds(1);
    digitalWrite(OE, HIGH);
    digitalWrite(CE, HIGH);

    return data;
}

void writeOp(int high, int low, int data)
{
    digitalWrite(CE, LOW);
    setAddress(high, low);
    delayMicroseconds(1);
    digitalWrite(WE, LOW);
    writeData(data);
    delayMicroseconds(1);
    digitalWrite(WE, HIGH);
    digitalWrite(CE, HIGH);
}

int readCommand(const char buffer[], int nread)
{
    int high, low;

    high = (fromHex(buffer[1]) << 8) + (fromHex(buffer[2]) << 4) + (fromHex(buffer[3]));
    low = (fromHex(buffer[4]) << 4) + (fromHex(buffer[5]));

    return readOp(high, low);
}

bool writeCommand(const char buffer[], int nread)
{
    int high, low, data;

    high = (fromHex(buffer[1]) << 8) + (fromHex(buffer[2]) << 4) + (fromHex(buffer[3]));
    low = (fromHex(buffer[4]) << 4) + (fromHex(buffer[5]));
    data = (fromHex(buffer[6]) << 4) + (fromHex(buffer[7]));

    writeOp(high, low, data);

    return true;
}

bool haveCompletion(int data, int written)
{
    return ((data & 0x80) == (written & 0x80));
}

bool haveFault(int data)
{
    return ((data & 0x20) != 0);
}

bool waitForCompletion(int high, int low, int written)
{
    int data;
    bool match;

    setDataIn();
    digitalWrite(CE, LOW);
    digitalWrite(OE, LOW);

    do
    {
        data = readOp(high, low);
        match = haveCompletion(data, written);
    } while (!match && !haveFault(data));

    if (!match)
    {
        data = readOp(high, low);
        match = haveCompletion(data, written);
    }

    setDataOut();
    digitalWrite(OE, HIGH);
    digitalWrite(CE, HIGH);
    return match;
}

bool reset()
{
    for (int i = 0; i < 3; i++)
    {
        writeCommand(RESET[i], 7);
    }
}

bool chipErase()
{
    for (int i = 0; i < 5; i++)
    {
        writeCommand(ERASE[i], 7);
    }
    writeOp(0x55, 0x55, 0x30);
    return waitForCompletion(0, 0, 0xff);
}

bool sectorErase(const char buffer[])
{
    int high;

    // 3-bit number shifted to A16...A14.
    // Thus, only sectors from 0 to 7 are supported.

    high = (fromHex(buffer[1]) << 6) & 0x1ff;

    for (int i = 0; i < 5; i++)
    {
        writeCommand(ERASE[i], 7);
    }
    writeOp(high, 0, 0x30);

    // Result of erase if 0xff written to all locations.

    return waitForCompletion(high, 0, 0xff);
}

bool program(const char buffer[], int nread)
{
    int high, low, data;

    high = (fromHex(buffer[1]) << 8) + (fromHex(buffer[2]) << 4) + (fromHex(buffer[3]));
    low = (fromHex(buffer[4]) << 4) + (fromHex(buffer[5]));
    data = (fromHex(buffer[6]) << 4) + (fromHex(buffer[7]));

    for (int i = 0; i < 3; i++)
    {
        writeCommand(PROGRAM[i], 7);
    }
    writeOp(high, low, data);

    return waitForCompletion(high, low, data);
}

void setup()
{
    Serial.begin(115200);
    pinMode(CE, OUTPUT);
    pinMode(OE, OUTPUT);
    pinMode(WE, OUTPUT);
    pinMode(A16, OUTPUT);

#ifdef CIRCUIT_74HC273
    pinMode(CS1, OUTPUT);
    pinMode(CS2, OUTPUT);
#endif

#ifdef CIRCUIT_74HC595
    pinMode(SHIFT_SCK, OUTPUT);
    pinMode(SHIFT_RCK, OUTPUT);
    pinMode(SHIFT_SI, OUTPUT);
#endif

    setDataOut();

    // Initial state for the flash device.

    digitalWrite(CE, HIGH);
    digitalWrite(OE, HIGH);
    digitalWrite(WE, HIGH);
    digitalWrite(A16, LOW);

    // Initial state for the latches.

#ifdef CIRCUIT_74HC273
    digitalWrite(CS1, LOW);
    digitalWrite(CS2, LOW);
#endif

#ifdef CIRCUIT_74HC595
    digitalWrite(SHIFT_SCK, LOW);
    digitalWrite(SHIFT_RCK, LOW);
#endif

    // Interface loop.

    Serial.println("?");
}

void loop()
{
    /* Read bytes, obtaining the number read excluding any newline terminator. */

    if (nread += Serial.readBytesUntil('\n', inbuffer + nread, BUFSIZE - nread))
    {
        /* Handle each command, waiting for the newline. */

        switch (inbuffer[0])
        {
            case 'r':
            reset();
            Serial.println("r");
            nread = 0;
            break;

            case 'R':
            if (nread >= 6)
            {
                Serial.println(readCommand(inbuffer, nread), HEX);
                nread = 0;
            }
            break;

            case 'W':
            if (nread >= 8)
            {
                if (writeCommand(inbuffer, nread))
                    Serial.println("W");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            case 'E':
            if (chipErase())
                Serial.println("E");
            else
                Serial.println("C");
            nread = 0;
            break;

            case 'S':
            if (nread >= 2)
            {
                if (sectorErase(inbuffer))
                    Serial.println("S");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            case 'P':
            if (nread >= 8)
            {
                if (program(inbuffer, nread))
                    Serial.println("P");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            default:
            Serial.println(inbuffer[0]);
            nread = 0;
            break;
        }

        Serial.flush();
    }
}

extern "C" void __cxa_pure_virtual(void) {
    while(1);
}

// tabstop=4 expandtab shiftwidth=4
